from isitfit.cost.ec2_analyze import pipeline_factory as ec2_cost_analyze
from isitfit.cost.ec2_optimize import pipeline_factory as ec2_cost_optimize

from isitfit.cost.redshift_analyze import pipeline_factory as redshift_cost_analyze
from isitfit.cost.redshift_optimize import pipeline_factory as redshift_cost_optimize

from isitfit.cost.account_cost_analyze  import pipeline_factory as account_cost_analyze
from isitfit.cost.account_cost_optimize import pipeline_factory as account_cost_optimize




def mig20191203a():
  """
  As of 0.17, rm ~/.isitfit/iterator_cache-*pkl
  """
  from isitfit.dotMan import DotMan
  import os
  import glob

  # rm ~/.isitfit/iterator_cache-*pkl
  # since that older path is deprecated in favor of /tmp/isitfit/iterator_cache*pkl
  del_p = os.path.join(DotMan().get_dotisitfit(), 'iterator_cache-*.pkl')
  del_p = os.path.expanduser(del_p) # do I need this?
  del_l = glob.glob(del_p)
  for del_i in del_l: os.unlink(del_i)


def mig20191203b():
  """
  As of 0.17, mv /tmp/isitfit*csv /tmp/isitfit/
  """
  from isitfit.dotMan import DotMan
  import tempfile
  import os
  import glob

  # rm /tmp/isitfit*csv
  # since started to use /tmp/isitfit/isitfit*csv
  del_p = os.path.join(tempfile.gettempdir(), 'isitfit*csv')
  new_tmpdir = DotMan().tempdir()
  del_l = glob.glob(del_p)
  for del_i in del_l:
    # os.unlink(del_i)
    new_i = os.path.basename(del_i)
    new_i = os.path.join(new_tmpdir, new_i)
    os.rename(del_i, new_i)


def mig20191203c():
  """
  As of 0.17, rm -rf /tmp/isitfit_ec2info.cache/
  """
  import tempfile
  import os
  import glob

  del_p = os.path.join(tempfile.gettempdir(), 'isitfit_ec2info.cache')
  import shutil
  shutil.rmtree(del_p, ignore_errors=True)



from isitfit.utils import logger, taglist2str


from isitfit.cost.redshift_common import CalculatorBaseRedshift
class CalculatorOptimizeRedshift(CalculatorBaseRedshift):

  def per_ec2(self, context_ec2):
      """
      # get all performance dataframes, on the cluster-aggregated level
      """

      # parent
      context_ec2 = super().per_ec2(context_ec2)

      # unpack
      rc_describe_entry = context_ec2['ec2_dict']
      df_single = context_ec2['df_single']

      # build string of taglist
      taglist = rc_describe_entry['Tags']
      taglist = taglist2str(taglist, context_ec2['filter_tags'])

      # summarize into maxmax, maxmin, minmax, minmin
      self.analyze_list.append({
        'Region': rc_describe_entry['Region'],
        'ClusterIdentifier': rc_describe_entry['ClusterIdentifier'],
        'NodeType': rc_describe_entry['NodeType'],
        'NumberOfNodes': rc_describe_entry['NumberOfNodes'],

        'CpuMaxMax': df_single.cpu_used_max.max(),
        #'CpuMaxMin': df_single.cpu_used_max.min(),
        #'CpuMinMax': df_single.cpu_used_min.max(),
        'CpuMinMin': df_single.cpu_used_min.min(),

        'tags': taglist
      })

      # done
      return context_ec2



  def calculate(self, context_all):
    if self.analyze_df.shape[0]==0:
      return context_all

    def classify_cluster_single(row):
        # classify
        if row.CpuMinMin > 70: return "Overused"
        if row.CpuMaxMax <  5: return "Idle"
        if row.CpuMaxMax < 30: return "Underused"
        return "Normal"

    # convert percentages to int since fractions are not very useful
    analyze_df = self.analyze_df

    analyze_df['classification'] = analyze_df.apply(classify_cluster_single, axis=1)

    return context_all




from isitfit.cost.base_reporter import ReporterBase
class ReporterOptimize(ReporterBase):
  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']
    analyze_df = self.analyzer.analyze_df

    # check if no data
    if analyze_df.shape[0]==0:
      return context_all

    # proceed
    analyze_df['CpuMaxMax'] = analyze_df['CpuMaxMax'].fillna(value=0).astype(int)
    analyze_df['CpuMinMin'] = analyze_df['CpuMinMin'].fillna(value=0).astype(int)

#    # copied from isitfit.cost.optimizationListener.storecsv...
#    # Update 2019-12-17 saving this table to csv is deprecated in favor of cost.account_cost_optimize.ServiceReporter.display2 which saves the data of both ec2 and redshift
#    import tempfile
#    from isitfit.dotMan import DotMan
#    with tempfile.NamedTemporaryFile(prefix='isitfit-full-redshift-', suffix='.csv', delete=False, dir=DotMan().tempdir()) as  csv_fh_final:
#      self.csv_fn_final = csv_fh_final.name
#      import click
#      from termcolor import colored
#      click.echo(colored("Saving final results to %s"%csv_fh_final.name, "cyan"))
#      analyze_df.to_csv(csv_fh_final.name, index=False)
#      click.echo(colored("Save complete", "cyan"))
#
#    # save in context for aggregator
#    context_all['csv_fn_final'] = self.csv_fn_final

    return context_all


# DEPRECATED
#  def display(self, context_all):
#    # copied from isitfit.cost.optimizationListener.display_all
#    analyze_df = self.analyzer.analyze_df
#
#    # display dataframe
#    from isitfit.utils import display_df
#    display_df(
#      "Redshift cluster classifications",
#      analyze_df,
#      self.csv_fn_final,
#      analyze_df.shape,
#      logger
#    )
#    return context_all


  def email(self, context_all):
      # silently return
      # raise Exception("Error emailing optimization: Not yet implemented")
      return context_all




def pipeline_factory(filter_region, ctx, filter_tags):
  # This is a factory method, so it doesn't make sense to display "Analyzing bla" if actually "foo" is analyzed first
  #logger.info("Optimizing redshift clusters")

  from .redshift_common import redshift_cost_core
  ra = CalculatorOptimizeRedshift()
  rr = ReporterOptimize()
  mm = redshift_cost_core(ra, rr, None, filter_region, ctx, filter_tags)

  # listener that was outed in the analyze step by the service aggregator
  # mm.add_listener('all', rr.display)

  return mm



from ...cost.metrics_datadog import DatadogManager, DatadogAssistant, DataNotFoundForHostInDdg, HostNotFoundInDdg, DatadogCached
import pandas as pd
import pytest


@pytest.mark.skip(reason="Can only test this with live credentials ATM. Need to mock")
def test_datadogman_1():
    ddg = DatadogManager()
    host_id='i-0f31feed76f7fb07c'
    df_all = ddg.get_metrics_all(host_id=host_id)
    assert df_all.shape[0] > 0
    #print(df_all.head())




from isitfit.utils import logger


import click

# Use "cls" to use the IsitfitCommand class to show the footer
# https://github.com/pallets/click/blob/8df9a6b2847b23de5c65dcb16f715a7691c60743/click/decorators.py#L92
from isitfit.cli.click_descendents import IsitfitCommand, isitfit_group, isitfit_option_base, isitfit_option_profile

from isitfit.utils import ping_matomo


@isitfit_group(help="Evaluate AWS EC2 costs", invoke_without_command=False)
@click.option('--filter-region', default=None, help='specify a single region against which to run cost analysis/optimization')
@isitfit_option_profile()
@isitfit_option_base(
  '--ndays',
  default=7,
  prompt='Number of days to lookback (between 1 and 90, use `isitfit cost --ndays=7 ...` to skip this prompt)',
  help='number of days to look back in the data history',
  type=click.IntRange(1, 90)
)
@click.pass_context
def cost(ctx, filter_region, ndays, profile):
  # FIXME click bug: `isitfit command subcommand --help` is calling the code in here. Workaround is to check --help and skip the whole section
  import sys
  if '--help' in sys.argv: return

  # gather anonymous usage statistics
  ping_matomo("/cost?filter_region=%s&ndays=%i"%(filter_region, ndays))

  # save to click context
  ctx.obj['ndays'] = ndays
  ctx.obj['filter_region'] = filter_region

  pass



# Check note above about ndays
# Another note about ndays: using click.IntRange for validation. Ref: https://click.palletsprojects.com/en/7.x/options/?highlight=prompt#range-options
@cost.command(help='Analyze AWS EC2 cost', cls=IsitfitCommand)
@click.option('--filter-tags', default=None, help='filter instances for only those carrying this value in the tag name or value')
@click.option('--save-details', is_flag=True, help='Save details behind calculations to CSV files')
@click.pass_context
def analyze(ctx, filter_tags, save_details):
    # gather anonymous usage statistics
    from isitfit.utils import b2l
    ping_matomo("/cost/analyze?filter_tags=%s&save_details=%s"%(filter_tags, b2l(save_details) ))

    # save to click context
    share_email = ctx.obj.get('share_email', [])

    #logger.info("Is it fit?")
    logger.info("Initializing...")


    # set up pipelines for ec2, redshift, and aggregator
    from isitfit.cost import ec2_cost_analyze, redshift_cost_analyze, account_cost_analyze
    mm_eca = ec2_cost_analyze(ctx, filter_tags, save_details)
    mm_rca = redshift_cost_analyze(share_email, filter_region=ctx.obj['filter_region'], ctx=ctx, filter_tags=filter_tags)

    # combine the 2 pipelines into a new pipeline
    mm_all = account_cost_analyze(mm_eca, mm_rca, ctx, share_email)

    # configure tqdm
    from isitfit.tqdmman import TqdmL2Quiet
    tqdml2 = TqdmL2Quiet(ctx)

    # Run pipeline
    mm_all.get_ifi(tqdml2)



@cost.command(help='Generate recommendations of optimal EC2 sizes', cls=IsitfitCommand)
@click.option('--n', default=-1, help='number of underused ec2 optimizations to find before stopping. Skip to get all optimizations')
@click.option('--filter-tags', default=None, help='filter instances for only those carrying this value in the tag name or value')
@click.pass_context
def optimize(ctx, n, filter_tags):
    # gather anonymous usage statistics
    ping_matomo("/cost/optimize?n=%i&filter_tags=%s"%(n, filter_tags ))

    # save to context
    share_email = ctx.obj.get('share_email', [])

    #logger.info("Is it fit?")
    logger.info("Initializing...")

    from isitfit.cost import ec2_cost_optimize, redshift_cost_optimize, account_cost_optimize
    mm_eco = ec2_cost_optimize(ctx, n, filter_tags)
    mm_rco = redshift_cost_optimize(filter_region=ctx.obj['filter_region'], ctx=ctx, filter_tags=filter_tags)

    # merge and run pipelines
    mm_all = account_cost_optimize(mm_eco, mm_rco, ctx)

    # configure tqdm
    from isitfit.tqdmman import TqdmL2Quiet
    tqdml2 = TqdmL2Quiet(ctx)

    # Run pipeline
    mm_all.get_ifi(tqdml2)



from .. import isitfit_version
import click

def version_core():
  print('isitfit version %s'%isitfit_version)


# Do not use the IsitfitCommand class to show the footer
# because the user will see the version info multiple times
@click.command(help="Show isitfit version")
def version():
  # gather anonymous usage statistics
  from isitfit.utils import ping_matomo
  ping_matomo("/version")

  version_core()
  return





# imports
import datetime as dt
from isitfit.utils import SECONDS_IN_ONE_DAY
import pandas as pd
import boto3

from isitfit.utils import logger, NoCloudwatchException


def raise_noCwExc(rc_id):
  raise NoCloudwatchException("No cloudwatch data for %s"%rc_id)


import numpy as np


class CloudwatchAssistant:
  """
  Manager for cloudwatch
  """

  def __init__(self):
    self.set_ndays(90) # default is 90 days


  def set_resource(self, region_name):
        """
        Easy-to-mock function since moto mock of cloudwatch is giving pagination error
        """
        boto3.setup_default_session(region_name = region_name)
        self.cloudwatch_resource = boto3.resource('cloudwatch')


  def id2iterator(self, rc_id, cloudwatch_namespace, entry_keyId):

    if cloudwatch_namespace is None:
      raise Exception("Derived class should set cloudwatch_namespace")

    metrics_iterator = self.cloudwatch_resource.metrics.filter(
        Namespace = cloudwatch_namespace,
        MetricName = 'CPUUtilization',
        Dimensions=[
            {'Name': entry_keyId, 'Value': rc_id},
        ]
      )
    return metrics_iterator


  def iterator2metric(self, metrics_iterator, rc_id):

    #logger.debug("redshift cluster details")
    #logger.debug(rc_describe_entry)

    for m_i in metrics_iterator:

        # skip node stats for now, and focus on cluster stats
        # i.e. dimensions only ClusterIdentifier, without the NodeID key
        if len(m_i.dimensions)>1:
          continue

        # exit the for loop and return this particular metric (cluster)
        return m_i

    # in case no cluster metrics found
    logger.debug("No cloudwatch metrics found for %s"%rc_id)
    raise_noCwExc(rc_id)


  def set_ndays(self, ndays):
    self.ndays = ndays

    # set start/end dates

    # FIXME? in mainManager, used pytz
    # dt_now_d=dt.datetime.now().replace(tzinfo=pytz.utc)
    dt_now_d = dt.datetime.utcnow()
    self.StartTime = dt_now_d - dt.timedelta(days=self.ndays)
    self.EndTime = dt_now_d

    # cloudwatch needs this to return the full data, including today?
    self.StartTime = self.StartTime.replace(hour=0, minute=0, second=0)
    self.EndTime = self.EndTime.replace(hour=23, minute=59, second=59)


  def metric2stats(self, metric):
    """
    For newly created instances, the Timestamp field is not reliable from here.
    It needs postprocessing by stats2df.
    For example, if today is 2019-12-17, an instance created today could return
    Timestamp=datetime.datetime(2019, 12, 13, 9, 0, tzinfo=tzutc())
    """
    logger.debug("fetch cw")
    logger.debug(metric.dimensions)

    # util func
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#CloudWatch.Metric.get_statistics
    # https://docs.aws.amazon.com/redshift/latest/mgmt/metrics-listing.html
    #
    # Note for redshift cluster
    # remember that max for redshift cluster = max of stats of all nodes
    response = metric.get_statistics(
        Dimensions=metric.dimensions,
        StartTime=self.StartTime,
        EndTime=self.EndTime,
        Period=SECONDS_IN_ONE_DAY,
        Statistics=['Minimum', 'Average', 'Maximum', 'SampleCount'],
        Unit = 'Percent'
    )
    logger.debug(response)
    return response


  def stats2df(self, response_metric, rc_id, ClusterCreateTime, cloudwatch_namespace):
    if len(response_metric['Datapoints'])==0:
      raise_noCwExc(rc_id)

    # convert to dataframe
    df = pd.DataFrame(response_metric['Datapoints'])

    # edit 2019-09-13: no need to subsample columns
    # The initial goal was to drop the "Unit" column (which just said "Percent"),
    # but it's not such a big deal, and avoiding this subsampling simplifies the code a bit
    # df = df[['Timestamp', 'SampleCount', 'Average']]

    # sort and append in case of multiple metrics
    df = df.sort_values(['Timestamp'], ascending=True)

    # before returning, convert dateutil timezone to pytz
    # for https://github.com/pandas-dev/pandas/issues/25423
    # via https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.dt.tz_convert.html
    # Edit 2019-09-25 Instead of keeping the full timestamp, just truncate to date, especially that this is just daily data
    # df['Timestamp'] = df.Timestamp.dt.tz_convert(pytz.utc)
    df['Timestamp'] = df.Timestamp.dt.date

    # cloudwatch bug: a newly created instance today will return a Timestamp before today
    # In this case, correcting the timestamp
    # Update 2019-12-17 This was fixed by setting the EndTime.hour,minute,second to the next midnight
    #if df.shape[0]==1:
    #  dt_now = dt.datetime.now().date()
    #  if ClusterCreateTime.date() == dt_now:
    #    if df.Timestamp.iloc[0] != dt_now:
    #      raise Exception("This cloudwatch bug was fixed by setting the hours/minutes/seconds of start/end time (location 1)")
    #      df.iloc[0, df.columns=='Timestamp'] = dt_now
    #
    ## drop points "before create time" (bug in cloudwatch?)
    ## Edit 2019-11-18 since this is daily data, and we don't really care about hours/minutes, just compare the y-m-d parts
    ## Update 2019-12-16 This is a weird bug
    #idx_cwbug = df['Timestamp'] >= ClusterCreateTime.date()
    #if not idx_cwbug.all():
    #  raise Exception("This cloudwatch bug was fixed by setting the hours/minutes/seconds of start/end time (location 2)")
    #  logger.debug("Cloudwatch bug of metric data after resource creation time: %s"%rc_id)

    #df = df[ idx_cwbug ]
    #if df.shape[0]==0: raise_noCwExc(rc_id)

    # calculate number of running hours
    # In the latest 90 days, sampling is per minute in cloudwatch
    # https://aws.amazon.com/cloudwatch/faqs/
    # Q: What is the minimum resolution for the data that Amazon CloudWatch receives and aggregates?
    # A: ... For example, if you request for 1-minute data for a day from 10 days ago, you will receive the 1440 data points ...
    if cloudwatch_namespace == 'AWS/EC2':
      df['nhours'] = np.ceil(df.SampleCount/60)
    elif cloudwatch_namespace == 'AWS/Redshift':
      # Redshift cloudwatch metrics are every 30 seconds (this seems to be the case by trial and error)
      # X points * 0.5 mins/point / 60 minutes/hr = Y hours
      df['nhours'] = np.ceil(df.SampleCount/60/2)

    # rename columns
    df.rename(columns={
        'Maximum': 'cpu_used_max',
        'Average': 'cpu_used_avg',
        'Minimum': 'cpu_used_min',
      },
      inplace=True
    )

    # append nan for memory
    df['ram_used_max'] = np.nan
    df['ram_used_avg'] = np.nan
    df['ram_used_min'] = np.nan

    logger.debug("returning dataframe.head")
    logger.debug(df.head())

    # print
    return df



class CloudwatchBase:
  cloudwatch_namespace = None
  entry_keyId = None

  def __init__(self):
    self.assistant = CloudwatchAssistant()

  def set_ndays(self, ndays):
    self.assistant.set_ndays(ndays)

  @property
  def ndays(self):
    """
    Expose the assistant.ndays member as a member here
    """
    return self.assistant.ndays

  def handle_main(self, rc_describe_entry, rc_id, rc_created):
    logger.debug("Fetching cloudwatch data for resource %s"%rc_id)

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#metric
    self.assistant.set_resource(region_name = rc_describe_entry['Region'])

    metrics_iterator = self.assistant.id2iterator(rc_id, self.cloudwatch_namespace, self.entry_keyId)

    # filter for 1 metric
    metric_single = self.assistant.iterator2metric(metrics_iterator, rc_id)
    response = self.assistant.metric2stats(metric_single)

    # dataframe of CPU Utilization, max and min, over 90 days
    df = self.assistant.stats2df(response, rc_id, rc_created, self.cloudwatch_namespace)

    return df


from .cacheManager import MetricCacheMixin


class CloudwatchCached(MetricCacheMixin, CloudwatchBase):
    """
    Manager for cloudwatch
    """

    def get_key(self, rc_id):
        # build key out of the same parameters in metric.get_statistics and metrics.filter
        #cache_key = "{Namespace}/{MetricName}/{DimensionName}/{DimensionValue}/{ndays}".format(
        #  Namespace = self.cloudwatch_namespace,
        #  MetricName = 'CPUUtilization',
        #  DimensionName = self.entry_keyId,
        #  DimensionValue = rc_id,
        #  ndays = self.ndays
        #)

        # KISS for now
        cache_key = "cloudwatch:cpu:%s:%i"%(rc_id, self.ndays)
        return cache_key

    def get_metrics_base(self, rc_describe_entry, rc_id, rc_created):
      return self.handle_main(rc_describe_entry, rc_id, rc_created)



class CloudwatchRedshift(CloudwatchCached):
  cloudwatch_namespace = 'AWS/Redshift'
  entry_keyId = 'ClusterIdentifier'


class CwRedshiftListener(CloudwatchRedshift):
  """
  Not yet deprecated
  AFAIK the datdog-redshift integration doesn't include memory metrics [1]
  So just referencing CwRedshiftListener in the redshift pipeline

  [1] https://docs.datadoghq.com/integrations/amazon_redshift/
  """
  def per_ec2(self, context_ec2):
        """
        Raises NoCloudwatchException if no data found in cloudwatch
        """
        rc_describe_entry, rc_id, rc_created = context_ec2['ec2_dict'], context_ec2['ec2_id'], context_ec2['ec2_launchtime']
        try:
          df_single = self.get_metrics_derived(rc_describe_entry, rc_id, rc_created)
          context_ec2['df_single'] = df_single
          return context_ec2
        except NoCloudwatchException:
          # Break chain of listeners
          # context_ec2['df_single'] = None
          return None




class CloudwatchEc2(CloudwatchCached):
  cloudwatch_namespace = 'AWS/EC2'
  entry_keyId = 'InstanceId'


class CwEc2Listener(CloudwatchEc2):
  """
  Deprecated in favor of metrics_automatic which references CloudwatchEc2 itself
  """

  def handle_main(self, ec2_obj):
    raise Exception("Deprecated")

    df_cw3 = super().handle_main({'Region': ec2_obj.region_name}, ec2_obj.instance_id, ec2_obj.launch_time)
    return df_cw3

  def per_ec2(self, context_ec2):
    """
    Raises NoCloudwatchException if no data found in cloudwatch
    """
    raise Exception("Deprecated")

    ec2_obj = context_ec2['ec2_obj']
    try:
      df_cw3 = self.handle_main(ec2_obj)
      context_ec2['df_metrics'] = df_cw3
      return context_ec2
    except NoCloudwatchException:
      # Break chain of listeners
      # context_ec2['df_metrics'] = None
      return None





import pytest
from isitfit.cli.click_descendents import IsitfitCliError

@pytest.fixture(scope='function')
def MockEmailManFactory(mocker):
    def get_class(response_list):
        class MockRequest:
          def __init__(self, response_list):
            self.counter = 0
            self.response_list = response_list
          def __call__(self, *args, **kwargs):
            print("mock.api.request")
            self.counter += 1
            if self.counter > len(self.response_list):
              raise ValueError("Called mock request %i times, and list of provided responses is of length %i"%(self.counter, len(self.response_list)))
            response_single = self.response_list[self.counter-1]
            return response_single, None

        mock_func = MockRequest(response_list)
        mocker.patch('isitfit.apiMan.ApiMan.request', side_effect=mock_func)
        mocker.patch('isitfit.apiMan.ApiMan.register', side_effect=lambda: None)

        # prepare
        from ..emailMan import EmailMan
        em = EmailMan(None, None, None)
        return em

    return get_class


def isitfit_return_helper(code, description, body):
  return {'isitfitapi_status': {'code': code, 'description': description}, 'isitfitapi_body': body}


@pytest.fixture
def sendemail_fac(MockEmailManFactory):
  def factory(response_list):
    # build a fake click command so that the click.prompt will be emulated
    # https://click.palletsprojects.com/en/7.x/testing/?highlight=test#input-streams
    import click
    @click.command()
    def cmd():
      em = MockEmailManFactory(response_list)
      em.send(['foo@bar.com'])

    return cmd

  return factory


class TestEmailManSendWithInputStream:
  """
  Note: If this test seems to hang forever, it might be the case that the click.prompt is looping forever
  given the new-line inputs below
  """

  def test_send_verificationInProgress_okAfter1stAttempt(self, sendemail_fac):
    r1 = isitfit_return_helper('Email verification in progress', 'foo', None)
    r2 = isitfit_return_helper('ok', 'foo', {'from': 'bla'})
    sendemail_cmd = sendemail_fac([r1,r2])

    # trigger
    from click.testing import CliRunner
    runner = CliRunner()
    result = runner.invoke(sendemail_cmd, input='\n')
    print(result.__dict__) # in case of exception, this will show details
    assert not result.exception
    assert True # no exception


  def test_send_verificationInProgress_failsAfter3Attempts(self, sendemail_fac):
    r1 = isitfit_return_helper('Email verification in progress', 'foo', None)
    sendemail_cmd = sendemail_fac([r1,r1,r1])

    # trigger
    from click.testing import CliRunner
    runner = CliRunner()
    result = runner.invoke(sendemail_cmd, input='\n\n\n')
    print(result.__dict__) # in case of exception, this will show details
    assert result.exception
    assert type(result.exception).__name__ == 'ValueError'



class TestEmailManSendNoInput:
  """
  These tests do not use TestEmailManSendWithInputStream as they do not reach a click.prompt, so they don't need to handle the stream
  """
  def test_send_empty(self, MockEmailManFactory):
    ret = isitfit_return_helper('error', 'foo', None)
    em = MockEmailManFactory([ret])

    # trigger
    em.send([])
    assert True # no exception, nothing happens


  def test_send_failErrorGeneral(self, MockEmailManFactory):
    ret = isitfit_return_helper('error', 'foo', None)
    em = MockEmailManFactory([ret])

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send(['bla@foo.bar'])


  def test_send_failNotOk(self, MockEmailManFactory):
    ret = isitfit_return_helper('hey', 'foo', None)
    em = MockEmailManFactory([ret])

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send(['bla@foo.bar'])


  def test_send_failSchema(self, MockEmailManFactory):
    ret = isitfit_return_helper('ok', 'foo', {})
    em = MockEmailManFactory([ret])

    # trigger
    with pytest.raises(IsitfitCliError) as e:
      em.send(['bla@foo.bar'])


  def test_send_ok(self, MockEmailManFactory):
    ret = isitfit_return_helper('ok', 'foo', {'from': 'bla'})
    em = MockEmailManFactory([ret])

    # no trigger
    em.send(['bla@foo.bar'])
    assert True # no exception







# Related
# https://docs.datadoghq.com/integrations/amazon_redshift/
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Paginator.DescribeClusters

from termcolor import colored
import click

from isitfit.utils import logger



class ReporterBase:
  def __init__(self):
    self.emailTo = None

  def postprocess(self, context_all):
    raise Exception("To be implemented by derived class")

  def display(self, context_all):
    raise Exception("To be implemented by derived class")

  def email(self, context_all):
      """
      ctx - click context
      """
      for fx in ['dataType', 'dataVal']:
        if not fx in context_all:
          raise Exception("Missing field from context: %s. This function should be implemented by the derived class"%fx)

      # unpack
      emailTo, ctx = context_all['emailTo'], context_all['click_ctx']

      # prompt user for email if not requested
      if not ctx.obj['skip_prompt_email']:
        from isitfit.utils import PromptToEmailIfNotRequested
        pte = PromptToEmailIfNotRequested()
        emailTo = pte.prompt(emailTo)

      # check if email requested
      if emailTo is None:
          return context_all

      if len(emailTo)==0:
          return context_all

      from isitfit.emailMan import EmailMan
      em = EmailMan(
        dataType=context_all['dataType'], # ec2, not redshift
        dataVal=context_all['dataVal'],
        ctx=ctx
      )
      em.send(emailTo)

      return context_all









"""
Functional tests run on shadi@autofitcloud.com AWS account
"""

from isitfit.cost.cloudtrail_iterator import *
import pandas as pd
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", None)

# variables for tests below
region_include = ['us-west-1', 'us-west-2', 'us-east-1', 'us-east-2']
from tqdm import tqdm as tqdmman

cache_man = None

EndTime = 'end time'
ec2_instances = [
  ( { 'Region': 'us-west-1',
      'ServiceName': 'ec2',
      'InstanceType': 't2.large',
      'InstanceId': 'i-1234'
    },
    'i-1234',
    None,
    None
  ),
  ( { 'Region': 'us-west-1',
      'ServiceName': 'redshift',
      'NodeType': 'dc2.large',
      'NumberOfNodes': 3,
      'ClusterIdentifier': 'test-1234'
    },
    'test-1234',
    None,
    None
  ),
]


import os 
BASEDIR = os.path.dirname(os.path.realpath(__file__))

# start of tests
def test_oneregion():
    # get all data and compare
    df_one = EventAggregatorOneRegion().get()
    assert not df_one.EventName.isnull().values.any()

    #print("")
    #print("one region")
    #print(df_one)

    fix_fn = "fixture_cloudtrailIterator_oneRegion_AutofitCloud-shadi_20191202.pkl"
    fix_fn = os.path.join(BASEDIR, fix_fn)
    # df_one.to_pickle(fix_fn)
    df_exp = pd.read_pickle(fix_fn)

    pd.testing.assert_frame_equal(df_exp, df_one)


def test_allregions():
    df_all = EventAggregatorAllRegions(region_include, tqdmman).get()

    print("")
    print("all regions")
    print(df_all)

    fix_fn = "fixture_cloudtrailIterator_allRegions_AutofitCloud-shadi_20191202.pkl"
    fix_fn = os.path.join(BASEDIR, fix_fn)
    #df_all.to_pickle(fix_fn)
    df_exp = pd.read_pickle(fix_fn)

    pd.testing.assert_frame_equal(df_exp, df_all)


def test_cached():
    # set up cache
    from isitfit.cost.cacheManager import RedisPandas
    cm = RedisPandas()

    # set up main class, and use a test cache key
    eac = EventAggregatorCached(region_include, tqdmman, cm)
    eac.cache_key = 'cloudtrail_iterator.test'

    # delete the key if used
    assert cm.isSetup()
    cm.connect()
    cm.redis_client.delete(eac.cache_key)

    # get pandas dataframe
    df_cached = eac.get()
    print("")
    print("cached")
    print(df_cached)

    # expect same file as un-cached
    fix_fn = "fixture_cloudtrailIterator_allRegions_AutofitCloud-shadi_20191202.pkl"
    fix_fn = os.path.join(BASEDIR, fix_fn)
    df_exp = pd.read_pickle(fix_fn)

    pd.testing.assert_frame_equal(df_exp, df_cached)



def test_postprocessed():
    df_post = EventAggregatorPostprocessed(region_include, tqdmman, cache_man, EndTime).get(ec2_instances, 2)

    print("")
    print("post-processed")
    print(df_post)

    fix_fn = "fixture_cloudtrailIterator_postProcessed_AutofitCloud-shadi_20191202.pkl"
    fix_fn = os.path.join(BASEDIR, fix_fn)
    #df_post.to_pickle(fix_fn)
    df_exp = pd.read_pickle(fix_fn)

    pd.testing.assert_frame_equal(df_exp, df_post)



import pytest
@pytest.fixture
def tqdml2_obj():
    class CtxWrap:
      obj = {'verbose': False, 'debug': False}

    from isitfit.tqdmman import TqdmL2Quiet
    gt = TqdmL2Quiet(
      CtxWrap()
    )
    return gt


class TestTqdmL2Quiet:
  
  def test_init(self, tqdml2_obj):
    x_iterator = tqdml2_obj([])
    x_list = list(x_iterator)
    assert True # no exception
    assert len(x_list)==0



from isitfit.utils import logger
import pandas as pd

from isitfit.cost.base_iterator import BaseIterator
class Ec2Iterator(BaseIterator):
  service_name = 'ec2'
  service_description = 'EC2 instances'
  paginator_name = 'describe_instances'
  # Notice that [] notation flattens the list of lists
  # http://jmespath.org/tutorial.html
  paginator_entryJmespath = 'Reservations[].Instances[]'
  paginator_exception = 'AuthFailure'
  entry_keyId = 'InstanceId'
  entry_keyCreated = 'LaunchTime'

  def __iter__(self):
    # over-ride the __iter__ to get the ec2 resource object for the current code (backwards compatibility)

    # method 1 for ec2
    # ec2_it = self.ec2_resource.instances.all()
    # return ec2_it

    # boto3 ec2 and cloudwatch data
    ec2_resource_all = {}
    import boto3

    # TODO cannot use directly use the iterator exposed in "ec2_it"
    # because it would return the dataframes from Cloudwatch,
    # whereas in the cloudwatch data fetch here, the data gets cached to redis.
    # Once the redshift.iterator can cache to redis, then the cloudwatch part here
    # can also be dropped, as well as using the "ec2_it" iterator directly
    # for ec2_dict in self.ec2_it:
    for ec2_dict, ec2_id, ec2_launchtime, _ in super().__iter__():
      if ec2_dict['Region'] not in ec2_resource_all.keys():
        boto3.setup_default_session(region_name = ec2_dict['Region'])
        ec2_resource_all[ec2_dict['Region']] = boto3.resource('ec2')

      ec2_resource_single = ec2_resource_all[ec2_dict['Region']]
      ec2_l = ec2_resource_single.instances.filter(InstanceIds=[ec2_dict['InstanceId']])
      ec2_l = list(ec2_l)
      if len(ec2_l)==0:
        continue # not found

      # yield first entry
      ec2_obj = ec2_l[0]
      ec2_obj.region_name = ec2_dict['Region']

      yield ec2_dict, ec2_id, ec2_launchtime, ec2_obj







from tabulate import tabulate

# https://pypi.org/project/termcolor/
from termcolor import colored



class CalculatorAnalyzeEc2:

  def __init__(self, ctx, save_details):
    # iterate over all ec2 instances
    self.sum_capacity = 0
    self.sum_used = 0
    self.df_all = []
    self.table = None # will contain the final table after calling `after_all`
    self.ctx = ctx

    # saving details to CSV
    self.save_details = save_details
    self.csv_fn_intermediate = None
    self.csv_fn_empty = True


  def handle_pre(self, context_pre):
    if not self.save_details: return context_pre
    import tempfile
    from isitfit.dotMan import DotMan
    csvi_prefix = 'isitfit-cost-analyze-ec2-details-1-'
    self.csv_fn_intermediate = tempfile.NamedTemporaryFile(prefix=csvi_prefix, suffix='.csv', delete=False, dir=DotMan().tempdir())
    return context_pre


  def per_ec2(self, context_ec2):
    """
    Listener function to be called upon the download of each EC2 instance's data
    ec2_obj - boto3 resource
    ec2_df - pandas dataframe with data from cloudwatch or datadog + cloudtrail + ec2instances.info catalog
    mm - mainManager class
    """
    # parse out context keys
    ec2_obj, ec2_df, mm = context_ec2['ec2_obj'], context_ec2['ec2_df'], context_ec2['mainManager']

    # results: 2 numbers: capacity (USD), used (USD)
    ec2_df['capacity_usd'] = ec2_df.nhours*ec2_df.cost_hourly
    res_capacity = ec2_df['capacity_usd'].sum()

    # use both the CPU Average from cloudwatch and the RAM average from datadog
    # If the RAM is nan, eg if data is from cloudwatch, the skipna=True ensures that this is calculated based on CPU only
    # >>> assert pd.DataFrame([{'a': 1, 'b': np.nan}]).mean(axis=1, skipna=True).iloc[0] == 1
    utilization_factor = ec2_df[['cpu_used_avg', 'ram_used_avg']].mean(axis=1, skipna=True)

    ec2_df['used_usd'] = ec2_df.nhours*ec2_df.cost_hourly*utilization_factor/100
    res_used   = ec2_df['used_usd'].sum()

    #logger.debug("res_capacity=%s, res_used=%s"%(res_capacity, res_used))

    self.sum_capacity += res_capacity
    self.sum_used += res_used
    self.df_all.append({'instance_id': ec2_obj.instance_id, 'capacity': res_capacity, 'used': res_used})

    # check if save details
    # http://stackoverflow.com/questions/17530542/ddg#17975690
    if self.save_details:
      ec2_df.to_csv(
        path_or_buf = self.csv_fn_intermediate.name,
        mode = 'w' if self.csv_fn_empty else 'a',
        header = self.csv_fn_empty,
        index = False
      )
      self.csv_fn_empty = False

    # save ec2_df again since added columns used and capacity
    context_ec2['ec2_df'] = ec2_df

    # done
    return context_ec2


  def after_all(self, context_all):
    # for debugging
    df_all = pd.DataFrame(self.df_all)
    logger.debug("\ncapacity/used per instance")
    logger.debug(df_all)
    logger.debug("\n")

    # set n analysed
    context_all['n_ec2_analysed'] = len(self.df_all)

    # dump to csv for details
    if self.save_details:
      import click

      # display message for first file
      csvi_desc ='Per ec2 and day'
      msg_info = "💾 Detail file 1/2: %s: %s"%(csvi_desc, self.csv_fn_intermediate.name)
      msg_info = colored(msg_info, "cyan")
      click.echo(msg_info)

      # save 2nd file and display message
      import tempfile
      from isitfit.dotMan import DotMan
      csvi_prefix = 'isitfit-cost-analyze-ec2-details-2-'
      csv_fh_final = tempfile.NamedTemporaryFile(prefix=csvi_prefix, suffix='.csv', delete=False, dir=DotMan().tempdir())

      df_all.to_csv(csv_fh_final.name, index=False)

      # display message about 2nd file
      csvi_desc = 'Per ec2 only   ' # 3 spaces just to align with "per ec2 and day
      msg_info = "💾 Detail file 2/2: %s: %s"%(csvi_desc, csv_fh_final.name)
      msg_info = colored(msg_info, "cyan")
      click.echo(msg_info)

      click.echo(colored("Consider viewing the CSVs in the terminal with visidata: `vd file.csv` (http://visidata.org/).", "cyan"))

      click.echo("") # empty breather line
    return context_all


from isitfit.utils import pd_series_frozenset_union, l2s
class BinCapUsed:
  def __init__(self):
    # sums, in a dataframe of time bins instead of 1 global number
    self.df_bins = None
    self.context_key = 'ec2_df'

  def _set_freq(self, ndays):
    # append x more month due to pandas date_range not yielding the EOM after dt_end
    # https://stackoverflow.com/a/4406260/4126114
    #from dateutil.relativedelta import relativedelta

    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

    if ndays <= 7:
      self.freq_start = '1D' # daily start
      self.freq_end = '1D' # daily end, note will require subtracting 1 day
      #self.freq_delta = relativedelta(days=1)
      return

    if ndays <= 30:
      self.freq_start = '1W-MON' # weekly start
      self.freq_end = '1W-SUN' # weekly end
      #self.freq_delta = relativedelta(days=7)
      return

    if ndays <= 60:
      self.freq_start = '1SMS' # semi-month start
      self.freq_end = '1SM' # semi-month end
      #self.freq_delta = relativedelta(days=15)
      return

    # otherwise monthly. Max data is 90 days from cloudwatch anyway
    self.freq_start = '1MS' # month start
    self.freq_end = '1M' # month end
    #self.freq_delta = relativedelta(months=1)
    return

  
  def do_resample_start(self, df_in):
    return df_in.resample(self.freq_start, label='left', closed='left')

  def do_resample_end(self, df_in):
    return df_in.resample(self.freq_end, label='right', closed='right')


#  def fix_resample_end(self, df_d2e, df_d2s):
#    # The problem showed up when computing min/max of Timestamp (i.e. dt_start, dt_end)
#    raise Exception("Shouldnt use self.fix_resample_end. It fixes the index but not the actual binning")
#
#    # some pandas issues with the freq_{start,end}
#    if self.freq_end=='1SM':
#      # FIXME bug in pandas: when freq_end='1SM', the mid-month start is e.g. 2019-10-15 and the previous mid-month end is also 2019-10-15.
#      # This is not the case for the mid-month ranges at the beginning, i.e. when start=2019-10-15 and end=2019-10-31
#      # This is inconsistent with the month-start and month-end which are 2019-10-01 and 2019-09-30 respectively
#      # So, substracting 1 day from the end date when it's the 15th
#      df_d2e = df_d2e.reset_index()
#      df_d2e['Timestamp'] = df_d2e.Timestamp.apply(lambda x: x - pd.DateOffset(1) if x.day==15 else x)
#      df_d2e.set_index('Timestamp', inplace=True)
#      return df_d2e
#
#    if self.freq_end=='1D':
#      # subtract 1 day from the end date since there was no freq_end='1DE' (day-end, hypothetically would yield the same date with hour:minute 23:59)
#      # such that [start,end] are inclusive just like the data
#      # Alternatively, just copy df_d2s
#      #df_d2e = df_d2e.reset_index()
#      #df_d2e['Timestamp'] = df_d2e.Timestamp.apply(lambda x: x - pd.DateOffset(1))
#      #df_d2e.set_index('Timestamp', inplace=True)
#      df_d2e = df_d2s.copy()
#      return df_d2e
#
#    # do nothing
#    return df_d2e

  def fix_resample_start(self, df_d2s, df_d2e, dt_start, dt_end):
    # FIXME bug in pandas: for ndays=30, hence freq_start='1W-MON', the df_d2s was getting 1 extra week for the week before the dt_start
    # I'm guessing that if the first day in df_d1 is a monday, the .resample adds 1 more entry for the monday before
    # Getting around this by chopping off dates before dt_start
    if df_d2s.shape[0]==(df_d2e.shape[0]+1):
      df_d2s = df_d2s[dt_start.date():dt_end.date()]

    return df_d2s


  def handle_pre(self, context_pre):
    # set freq
    ndays = context_pre['mainManager'].ndays
    self._set_freq(ndays)

    # util vars
    dt_start = context_pre['mainManager'].StartTime
    dt_end   = context_pre['mainManager'].EndTime

    # set df_bins
    dt_daily = pd.date_range(start=dt_start.date(), end=dt_end.date(),  freq='1D')
    df_d1 = pd.DataFrame({
      'Timestamp': dt_daily,
      'dummy': [0]*len(dt_daily),
    })
    df_d1.set_index('Timestamp', inplace=True)
    df_d2e = self.do_resample_end(df_d1).sum()
    df_d2s = self.do_resample_start(df_d1).sum()
    #df_d2e = self.fix_resample_end(df_d2e, df_d2s)
    #df_d2s = self.fix_resample_start(df_d2s, df_d2e, dt_start, dt_end)

    # append 1 more month due to pandas date_range not yielding the EOM after dt_end
    # Update 2019-12-09 no longer needed due to usage of .resample instead of .date_range
    #dt_start2 = dt_start - self.freq_delta
    #dt_end2   = dt_end   + self.freq_delta

    # get list of dates with freq
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.date_range.html
    # Update 2019-12-09 instead of using date_range to manually build the dataframe,
    # just use .resample on a dummy dataframe and pick up the index
    #dt_range_start = pd.date_range(start=dt_start2.date(), end=dt_end.date(),  freq=self.freq_start)
    #dt_range_end   = pd.date_range(start=dt_start.date(),  end=dt_end2.date(), freq=self.freq_end  )

    # Do not convert to `.dt.date` yet to avoid losing the `.dt` accessor later, eg `.dt.strftime`
    dt_range_start = df_d2s.reset_index().Timestamp.tolist()
    dt_range_end   = df_d2e.reset_index().Timestamp.tolist()

    # create dataframe
    # Alternatively than creating lists of dates dt_range_{start,end}, could have merged as follows:
    # df_d2s.reset_index().merge(df_d2e.reset_index(), left_index=True, right_index=True, how='outer')
    self.df_bins = pd.DataFrame({
      'Timestamp': dt_range_end, # use dt_range_end in conjunction with label='right' in the .resample calls below
      'capacity_usd': 0,
      'used_usd': 0,
      'count_analyzed': 0,
      'regions_set': [frozenset([]) for x in dt_range_end], # list of sets
      'dt_start': dt_range_end, # init at end, then search for min in per_ec2
      'dt_end': dt_range_start, # init at start, then search for max in per_ec2
    })
    self.df_bins.set_index('Timestamp', inplace=True)

    return context_pre


  def per_ec2(self, context_ec2):
    if self.df_bins is None:
      raise Exception("Call handle_pre first to set the dataframe")

    ec2_df = context_ec2[self.context_key]
    df_add = ec2_df[['Timestamp', 'capacity_usd', 'used_usd']].copy()

    # index + add a duplicate column for the timestamp and use it to get min/max dates in a time bin
    df_add['Timestamp'] = pd.to_datetime(df_add['Timestamp'])
    df_add['ts2'] = df_add.Timestamp.tolist()
    df_add.set_index('Timestamp', inplace=True)

    # resample ints
    df_me = self.do_resample_end(df_add[['capacity_usd', 'used_usd']]).sum()
    #df_ignore = self.do_resample_start(df_add[['capacity_usd', 'used_usd']]).sum()
    #df_me = self.fix_resample_end(df_me, df_ignore)

    # util vars
    dt_start = context_ec2['mainManager'].StartTime
    dt_end   = context_ec2['mainManager'].EndTime

    # resample dates. Note that min/max doesn't matter below for freq_start='1D'.
    # Intentionally calling do_resample_end for both dfme_{end,start} so that they get the same Timeindex
    #dfme_end_max = self.do_resample_end(df_add['ts2']).max()
    #dfme_end_min = self.do_resample_end(df_add['ts2']).min()
    #dfme_start_max = self.do_resample_start(df_add['ts2']).max()
    #dfme_start_min = self.do_resample_start(df_add['ts2']).min()
    #df_me['dt_end'  ] = self.fix_resample_end(  dfme_end_max, dfme_start_max)
    #df_me['dt_start'] = self.fix_resample_end(  dfme_end_min, dfme_start_min)
    df_me['dt_end'  ] = self.do_resample_end(df_add['ts2']).max()
    df_me['dt_start'] = self.do_resample_end(df_add['ts2']).min()


    # dummy column showing 1 for the current instance, ie where there is any capacity
    # df_me['count_analyzed'] = 1
    df_me['count_analyzed'] = (df_me.capacity_usd > 0).astype(int)

    # append region
    #df_me['region'] = set([context_ec2['ec2_dict']['Region'])
    df_me['regions_set'] = df_me.apply(lambda row: frozenset([context_ec2['ec2_dict']['Region']]), axis=1)

    # cast all to int for simplicity
    for fx in ['capacity_usd', 'used_usd']:
      df_me[fx] = df_me[fx].fillna(value=0)
      df_me[fx] = df_me[fx].astype(int)

    # Add dataframes ints
    # Using the "+" operator will just fill missing indeces with NaN
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.add.html
    add_cols = ['capacity_usd', 'used_usd', 'count_analyzed']
    df_sub1 = self.df_bins[add_cols]
    df_sub2 = df_me[add_cols]
    df_sub1 = df_sub1.add(df_sub2, fill_value=0)
    self.df_bins[add_cols] = df_sub1

    # bug in pandas.DataFrame.add function: it converts the int to float 
    # This is despite the dtypes of df_me and self.df_bins being int
    # and the fill_value being int
    # Probably due to missing index values feeding in "na" and later internally replacing the na with 0
    # TODO file issue and/or PR?
    for fx in ['capacity_usd', 'used_usd', 'count_analyzed']:
      self.df_bins[fx] = self.df_bins[fx].astype(int)

    # add the region sets
    self.df_bins['regions_set'] = pd_series_frozenset_union(self.df_bins['regions_set'], df_me['regions_set'])

    # add the start/end dates
    self.df_bins['dt_start'] = pd.concat([self.df_bins.dt_start, df_me.dt_start], axis=1).min(axis=1)
    self.df_bins['dt_end'  ] = pd.concat([self.df_bins.dt_end,   df_me.dt_end  ], axis=1).max(axis=1)

    # done
    return context_ec2

  def after_all(self, context_all):
    # add col for utilization in percentage
    def calc_usedPct(row):
      if row.capacity_usd==0: return 0
      o = row.used_usd / row.capacity_usd * 100
      return int(o)

    self.df_bins['used_pct'] = self.df_bins.apply(calc_usedPct, axis=1)

    # add column for regions as string
    self.df_bins['regions_str'] = self.df_bins['regions_set'].apply(lambda x: "0" if len(x)==0 else "%i (%s)"%(len(x), l2s(x)))

    # cases where dt_start > dt_end are those where there was no data and the initialization remained
    # so overwrite with na
    # Update 2019-12-11 Now that the df_bins timestamps are set with resample and dt_end is inclusive,
    # instead of setting to na, just swap the start/end fake timestamps which represent the end/start of the periods
    #import numpy as np
    #self.df_bins['dt_start'] = self.df_bins.apply(lambda row: np.nan if row.count_analyzed==0 else row.dt_start, axis=1)
    #self.df_bins['dt_end']   = self.df_bins.apply(lambda row: np.nan if row.count_analyzed==0 else row.dt_end  , axis=1)
    self.df_bins['dt_start_bkp']  = self.df_bins['dt_start']
    self.df_bins['dt_start'] = self.df_bins.apply(lambda row: row.dt_end       if row.count_analyzed==0 else row.dt_start, axis=1)
    self.df_bins['dt_end']   = self.df_bins.apply(lambda row: row.dt_start_bkp if row.count_analyzed==0 else row.dt_end  , axis=1)
    del self.df_bins['dt_start_bkp']

    # convert the dt_{start,end} back to dates again, given the nans
    for fx in ['dt_start', 'dt_end']: self.df_bins[fx] = pd.to_datetime(self.df_bins[fx])

    # Bugfix for cloudwatch:
    # When 90>=ndays>=64, cloudwatchman/metric.get_statistics returns data with max Timestamp on mainmanager.EndTime
    # Otherwise, the data max Timestamp is EndTime - 1 day
    # Here, get around this problem by incrementing by 1 day (or just set to mainmanager.EndTime)
    # This was tested on 2019-12-10 10:00 am UTC, and the last date was Dec 9 for ndays<64 and Dec 10 for ndays>=64
    if pd.notnull(self.df_bins['dt_end'].iloc[-1]):
      dt_max = context_all['mainManager'].EndTime.date()
      import datetime as dt
      dt_lastp1 = self.df_bins.dt_end.iloc[-1].date() + dt.timedelta(days=1)
      self.df_bins.iloc[-1, self.df_bins.columns=='dt_end'] = min(dt_max, dt_lastp1)

    # inject result for reporter access
    context_all['df_bins'] = self.df_bins
    return context_all


from isitfit.cost.base_reporter import ReporterBase

class ReporterAnalyzeEc2(ReporterBase):

  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']
    n_ec2_total, self.mm, n_ec2_analysed, region_include = context_all['n_ec2_total'], context_all['mainManager'], context_all['n_ec2_analysed'], context_all['region_include']

    # proceed
    cwau_val = 0
    if self.analyzer.sum_capacity!=0:
      cwau_val = self.analyzer.sum_used/self.analyzer.sum_capacity*100

    cwau_color = 'yellow'
    if cwau_val >= 70:
      cwau_color = 'green'
    elif cwau_val <= 30:
      cwau_color = 'red'

    dt_start = self.mm.StartTime.strftime("%Y-%m-%d")
    dt_end   = self.mm.EndTime.strftime("%Y-%m-%d")

    ri_max = 3
    ri_ell = "" if len(region_include)<=ri_max else "..."
    ri_str = ", ".join(region_include[:ri_max])+ri_ell
    
    self.table = [
            {'color': '',
             'label': "Start date",
             'value': "%s"%dt_start
            },
            {'color': '',
             'label': "End date",
             'value': "%s"%dt_end
            },
            {'color': '',
             'label': "Regions",
             'value': "%i (%s)"%(len(region_include), ri_str)
            },
            {'color': '',
             'label': "EC2 machines (total)",
             'value': "%i"%n_ec2_total
            },
            {'color': '',
             'label': "EC2 machines (analyzed)",
             'value': "%i"%n_ec2_analysed
            },
            {'color': 'cyan',
             'label': "Billed cost",
             'value': "%0.0f $"%self.analyzer.sum_capacity
            },
            {'color': 'cyan',
             'label': "Used cost",
             'value': "%0.0f $"%self.analyzer.sum_used
            },
            {'color': cwau_color,
             'label': "CWAU (Used/Billed)",
             'value': "%0.0f %%"%cwau_val
            },
    ]

    # save in context for aggregator
    context_all['table'] = self.table

    # done
    return context_all


  def display(self, context_all):
    def get_row(row):
        def get_cell(i):
          retc = row[i] if not row['color'] else colored(row[i], row['color'])
          return retc
        
        retr = [get_cell('label'), get_cell('value')]
        return retr

    dis_tab = [get_row(row) for row in self.table]

    # logger.info("Summary:")
    import click
    click.echo("Cost-Weighted Average Utilization (CWAU) of the AWS EC2 account:")
    click.echo("")
    click.echo(tabulate(dis_tab, headers=['Field', 'Value']))
    click.echo("")
    click.echo("For reference:")
    click.echo(colored("* CWAU >= 70% is well optimized", 'green'))
    click.echo(colored("* CWAU <= 30% is underused", 'red'))

    return context_all


  def email(self, context_all):
      """
      ctx - click context
      """
      context_2 = {}
      context_2['emailTo'] = context_all['emailTo']
      context_2['click_ctx'] = context_all['click_ctx']
      context_2['dataType'] = 'cost analyze' # ec2, not redshift
      context_2['dataVal'] = {'table': self.table}
      super().email(context_2)

      return context_all





def pipeline_factory(ctx, filter_tags, save_details):
    # moved these imports from outside the function to inside it so that `isitfit --version` wouldn't take 5 seconds due to the loading
    from isitfit.cost.mainManager import MainManager
    from isitfit.cost.cloudtrail_ec2type import CloudtrailCached

    # manager of redis-pandas caching
    from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
    cache_man = RedisPandasCacheManager()

    # 2019-12-16 Deprecate the datadog and cloudwatch listeners in favor of the automatic fallback listener
    # from isitfit.cost.metrics_datadog import DatadogListener
    # from isitfit.cost.metrics_cloudwatch import CwEc2Listener
    from isitfit.cost.metrics_datadog import DatadogCached
    from isitfit.cost.metrics_cloudwatch import CloudwatchEc2
    from isitfit.cost.metrics_automatic import MetricsListener
    ddg = DatadogCached(cache_man)
    cloudwatchman = CloudwatchEc2(cache_man)
    metrics = MetricsListener(ddg, cloudwatchman)
    metrics.set_ndays(ctx.obj['ndays'])

    from isitfit.cost.ec2_common import Ec2TagFilter
    from isitfit.cost.catalog_ec2 import Ec2Catalog
    from isitfit.cost.ec2_common import Ec2Common

    from isitfit.tqdmman import TqdmL2Verbose
    tqdml2_obj = TqdmL2Verbose(ctx)

    share_email = ctx.obj.get('share_email', None)
    ul = CalculatorAnalyzeEc2(ctx, save_details)



    etf = Ec2TagFilter(filter_tags)


    ra = ReporterAnalyzeEc2()

    mm = MainManager("EC2 cost analyze", ctx)
    mm.set_ndays(ctx.obj['ndays'])

    ec2_cat = Ec2Catalog()
    ec2_common = Ec2Common()
    ec2_it = Ec2Iterator(ctx.obj['filter_region'], tqdml2_obj)

    # boto3 cloudtrail data
    cloudtrail_manager = CloudtrailCached(mm.EndTime, cache_man, tqdml2_obj)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_email_in_context = lambda context_all: dict({'emailTo': share_email}, **context_all)
    inject_analyzer = lambda context_all: dict({'analyzer': ul}, **context_all)

    # binning
    bcs = BinCapUsed()

    # utilization listeners
    mm.set_iterator(ec2_it)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('pre', cloudtrail_manager.init_data)
    mm.add_listener('pre', ec2_cat.handle_pre)
    mm.add_listener('pre', ul.handle_pre)
    mm.add_listener('pre', bcs.handle_pre)
    mm.add_listener('ec2', etf.per_ec2)
    mm.add_listener('ec2', metrics.per_host)
    mm.add_listener('ec2', cloudtrail_manager.single)
    mm.add_listener('ec2', ec2_common._handle_ec2obj)
    mm.add_listener('ec2', ul.per_ec2)
    mm.add_listener('ec2', bcs.per_ec2)
    mm.add_listener('all', metrics.display_status)
    mm.add_listener('all', ec2_common.after_all)
    mm.add_listener('all', ul.after_all)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', ra.postprocess)
    mm.add_listener('all', bcs.after_all)
    #mm.add_listener('all', ra.display)
    #mm.add_listener('all', inject_email_in_context)
    #mm.add_listener('all', ra.email)

    return mm





# redshift pricing as of 2019-11-12 in USD per hour, on-demand, ohio
# https://aws.amazon.com/redshift/pricing/
from isitfit.cost.catalog_redshift import redshiftPricing_dict



from isitfit.cost.base_iterator import BaseIterator
class RedshiftPerformanceIterator(BaseIterator):
  service_name = 'redshift'
  service_description = 'Redshift clusters'
  paginator_name = 'describe_clusters'
  paginator_entryJmespath = 'Clusters[]'
  paginator_exception = 'InvalidClientTokenId'
  entry_keyId = 'ClusterIdentifier'
  entry_keyCreated = 'ClusterCreateTime'




# AWS_DEFAULT_REGION=us-east-2 python3 -m isitfit.cost.test_redshift
# Related
# https://docs.datadoghq.com/integrations/amazon_redshift/
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Paginator.DescribeClusters

import pandas as pd
from isitfit.cost.metrics_cloudwatch import NoCloudwatchException



def tagsContain(f_tn, ec2_dict):
  """
  Similar to isitfit.cost.ec2_common.tagsContain
  """
  if ec2_dict['Tags'] is None: return False
  if len(ec2_dict['Tags']): return False

  for t in ec2_dict['Tags']:
    for k in ['Key', 'Value']:
      if f_tn in t[k].lower():
        return True
  
  return False


class RedshiftTagFilter:
  """
  Similar to isitfit.cost.ec2_common.Ec2TagFilter
  """
  def __init__(self, filter_tags):
    self.filter_tags = filter_tags

  def per_cluster(self, context_cluster):
    # if filters requested, check that this instance passes

    # set in context for the sake of filtering the taglist in redshift_optimize.Calculator.per_ec2
    context_cluster['filter_tags'] = self.filter_tags

    if self.filter_tags is None:
      # to continue with other listeners
      return context_cluster

    f_tn = self.filter_tags.lower()
    passesFilter = tagsContain(f_tn, context_cluster['ec2_dict'])

    if not passesFilter:
      # break other listeners
      return None

    # otherwise continue
    return context_cluster



class CalculatorBaseRedshift:


  def __init__(self):
    # define the list in the constructor because if I define it as a class member above,
    # then it gets reused between instantiations of derived classes
    self.analyze_list = []
    self.analyze_df = None


  def per_ec2(self, context_ec2):
      rc_describe_entry = context_ec2['ec2_dict']

      # for types not yet in pricing dictionary above
      rc_type = rc_describe_entry['NodeType']
      if rc_type not in redshiftPricing_dict.keys():
        raise NoCloudwatchException

      return context_ec2


  def after_all(self, context_all):
    # To be used by derived class *after* its own implementation

    # gather into a single dataframe
    self.analyze_df = pd.DataFrame(self.analyze_list)

    # update number of analyzed clusters
    context_all['n_rc_analysed'] = self.analyze_df.shape[0]

    # Edit 2019-11-20 no need to through exception here
    # This way, the code can proceed to show a report, and possibly proceed to other services than redshift
    #if context_all['n_rc_analysed']==0:
    #  from isitfit.cli.click_descendents import IsitfitCliError
    #  raise IsitfitCliError("No redshift clusters analyzed", context_all['click_ctx'])

    return context_all


  def calculate(self, context_all):
    raise Exception("To be implemented by derived class")







# Related
# https://docs.datadoghq.com/integrations/amazon_redshift/
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift.html#Redshift.Paginator.DescribeClusters

import click

from isitfit.utils import logger



def redshift_cost_core(ra, rr, share_email, filter_region, ctx, filter_tags):
    """
    ra - Analyzer
    rr - Reporter
    """

    # data layer
    from isitfit.tqdmman import TqdmL2Verbose
    tqdmman = TqdmL2Verbose(ctx)

    ri = RedshiftPerformanceIterator(filter_region, tqdmman)

    # pipeline
    from isitfit.cost.mainManager import MainManager
    from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
    from isitfit.cost.metrics_cloudwatch import CwRedshiftListener
    from isitfit.cost.ec2_common import Ec2Common
    from isitfit.cost.cloudtrail_ec2type import CloudtrailCached

    mm = MainManager("Redshift cost analyze or optimize", ctx)
    mm.set_ndays(ctx.obj['ndays'])

    cache_man = RedisPandasCacheManager()

    # manager of cloudwatch
    cwman = CwRedshiftListener(cache_man)
    cwman.set_ndays(ctx.obj['ndays'])

    # common stuff
    ec2_common = Ec2Common()

    # boto3 cloudtrail data
    # FIXME note that if two pipelines are run, one for ec2 and one for redshift, then this Object fetches the same data twice
    # because the base class behind it does both ec2+redshift at once
    # in the init_data phase
    cloudtrail_manager = CloudtrailCached(mm.EndTime, cache_man, tqdmman)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_analyzer = lambda context_all: dict({'analyzer': ra}, **context_all)

    # binning
    do_binning = False
    ra_name = type(ra).__name__
    if ra_name == 'CalculatorOptimizeRedshift':
      do_binning = False
    elif ra_name == 'CalculatorAnalyzeRedshift':
      do_binning = True
    else:
      raise Exception("Invalid calculator class passed: %s"%ra_name)

    if do_binning:
      from isitfit.cost.ec2_analyze import BinCapUsed
      bcs = BinCapUsed()
      bcs.context_key = 'df_single'

    # setup pipeline
    mm.set_iterator(ri)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('pre', cloudtrail_manager.init_data)

    if do_binning:
      mm.add_listener('pre', bcs.handle_pre)

    rtf = RedshiftTagFilter(filter_tags)
    mm.add_listener('ec2', rtf.per_cluster)

    mm.add_listener('ec2', cwman.per_ec2)
    mm.add_listener('ec2', cloudtrail_manager.single)
    mm.add_listener('ec2', ra.per_ec2)

    if do_binning:
      mm.add_listener('ec2', bcs.per_ec2)

    mm.add_listener('all', ec2_common.after_all) # just show IDs missing cloudwatch/cloudtrail
    mm.add_listener('all', ra.after_all)
    mm.add_listener('all', ra.calculate)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', rr.postprocess)

    if do_binning:
      mm.add_listener('all', bcs.after_all)

    #inject_email_in_context = lambda context_all: dict({'emailTo': share_email}, **context_all)
    #mm.add_listener('all', rr.display)
    #mm.add_listener('all', inject_email_in_context)
    #mm.add_listener('all', rr.email)

    return mm






import pytest


# mocker fixture becomes available after installing pytest-mock
# https://github.com/pytest-dev/pytest-mock
def test_pingMatomo_unit(mocker):
  from ..utils import ping_matomo
  def mockreturn(url, json, timeout): return "foo"
  mocked_post = mocker.patch('isitfit.utils.requests.post', side_effect=mockreturn)
  ping_matomo("/test")

  # check that mocked object is called
  # https://github.com/pytest-dev/pytest-mock/commit/68868872195135bdb90d45a5cb0d609400943eae
  mocked_post.assert_called()



def test_pingMatomo_functional(mocker):
  from ..utils import ping_matomo
  ping_matomo("/test")





def test_isitfitCliError():
    from isitfit.cli.click_descendents import IsitfitCliError

    class MockContext:
        obj = {'bar': 1}
        command = None

    ctx = MockContext()
    with pytest.raises(IsitfitCliError) as e:
        raise IsitfitCliError("foo", ctx)



def test_mergeSeriesOnTimestampRange():
  import pandas as pd
  df_cpu = pd.DataFrame({'Timestamp': [1,2,3,4], 'field_1': [5,6,7,8]})
  df_type = pd.DataFrame({'Timestamp': [1,3], 'field_2': ['a','b']})

  # update 2019-11-20 had initially written example as field_2: a, a, b, b
  # but maybe that was an outdated example
  expected = pd.DataFrame({'Timestamp': [1,2,3,4], 'field_1': [5,6,7,8], 'field_2': ['a','b','b','b']})

  # reverse sort
  df_cpu  = df_cpu.sort_values(['Timestamp'], ascending=False)
  df_type = df_type.sort_values(['Timestamp'], ascending=False)

  # set index
  df_type = df_type.set_index('Timestamp')

  # test
  from ..utils import mergeSeriesOnTimestampRange
  actual = mergeSeriesOnTimestampRange(df_cpu, df_type, ['field_2'])

  # straight sort
  actual = actual.sort_values(['Timestamp'], ascending=True)

  #print(expected)
  #print(actual)
  pd.testing.assert_frame_equal(expected, actual)


def test_b2l():
  from isitfit.utils import b2l
  a = b2l(True)
  assert a=='T'
  a = b2l(False)
  assert a=='F'



def test_l2s():
  from isitfit.utils import l2s
  a = l2s(list(range(10)))
  assert a=='0,1,...,8,9'


# https://stackoverflow.com/a/40884671/4126114
@pytest.mark.parametrize("filter_val,expect_out", [(None, 'app = isitfit\napp = another'), ('is', 'app = isitfit'), ('foo', '')])
def test_taglist2str(filter_val, expect_out):
  from isitfit.utils import taglist2str

  a = taglist2str([{'Key':'app', 'Value':'isitfit'}, {'Key':'app', 'Value':'another'}], filter_val)
  assert a == expect_out



def test_pandasSets_sameIndex():
  from isitfit.utils import pd_series_frozenset_union
  import pandas as pd

  fset = frozenset
  s1=pd.Series([fset([1]), fset([1,2])])
  s2=pd.Series([fset([1]), fset([1,3])])
  a = pd_series_frozenset_union(s1, s2)
  e = pd.DataFrame({'a3': [fset([1]), fset([1,2,3])]})
  pd.testing.assert_series_equal(a, e.a3)


def test_pandasSets_differentIndex():
  from isitfit.utils import pd_series_frozenset_union
  import pandas as pd

  fset = frozenset
  s1=pd.Series([fset([1]), fset([1,2])], index=[0,1])
  s2=pd.Series([fset([1])], index=[0])
  actual = pd_series_frozenset_union(s1, s2)
  expected = pd.DataFrame({'a3': [fset([1]), fset([1,2])]})
  pd.testing.assert_series_equal(actual, expected.a3)



from isitfit.utils import AwsProfileMan
class TestAwsProfileMan:
  def test_init(self):
    pm = AwsProfileMan()
    assert True # no exception

  def test_validateProfile(self):
    class MockClickCtx:
      obj = {
      }

    click_ctx = MockClickCtx()

    pm = AwsProfileMan()
    actual = pm.validate_profile(click_ctx, "profile", "default")
    assert actual=="default"

    import click
    with pytest.raises(click.BadParameter):
      pm.validate_profile(click_ctx, "profile", "inexistant")


  def test_prompt(self):
    pm = AwsProfileMan()
    x = pm.prompt()
    assert x is not None



# get fixture
from isitfit.tests.cli.test_clickDescendents import ping_matomo

# test functions using above fixture
class TestPromptToEmailIfNotRequested:

  def test_ok(self, ping_matomo):
    """
    No need for click.command here as this will never reach the click.prompt line
    """
    from isitfit.utils import PromptToEmailIfNotRequested
    pte = PromptToEmailIfNotRequested()
    import tempfile
    with tempfile.NamedTemporaryFile() as fh:
      pte.last_email_cl.fn = fh.name # overwrite file to save last-used email
      expected = 'me@example.com'
      actual = pte.prompt(expected)
      assert actual == expected


  def test_oneInputNoLast(self, ping_matomo):
    """
    # build a fake click command so that the click.prompt will be emulated
    # https://click.palletsprojects.com/en/7.x/testing/?highlight=test#input-streams
    """
    import click
    @click.command()
    def cmd():
      from isitfit.utils import PromptToEmailIfNotRequested
      pte = PromptToEmailIfNotRequested()
      import tempfile
      with tempfile.NamedTemporaryFile() as fh:
        pte.last_email_cl.fn = fh.name # overwrite file to save last-used email
        pte.prompt(None)

    # trigger
    from click.testing import CliRunner
    runner = CliRunner()
    result = runner.invoke(cmd, input='me@example.com\n')
    print(result.__dict__) # in case of exception, this will show details
    assert not result.exception
    assert '[skip]' in result.output


  def test_oneInputSetLast(self, ping_matomo):
    """
    # build a fake click command so that the click.prompt will be emulated
    # https://click.palletsprojects.com/en/7.x/testing/?highlight=test#input-streams
    """
    import click
    @click.command()
    def cmd():
      from isitfit.utils import PromptToEmailIfNotRequested
      pte = PromptToEmailIfNotRequested()
      import tempfile
      with tempfile.NamedTemporaryFile() as fh:
        pte.last_email_cl.fn = fh.name # overwrite file to save last-used email
        pte.last_email_cl.set('me@example.com')
        pte.prompt(None)

    # trigger
    from click.testing import CliRunner
    runner = CliRunner()
    result = runner.invoke(cmd, input='\n')
    print(result.__dict__) # in case of exception, this will show details
    assert not result.exception
    assert '[skip]' not in result.output
    assert '[me@example.com]' in result.output


  def test_userInput(self):
    import click
    from click.testing import CliRunner

    class MyWrap:
      def dummyFac(self, emailIn, emailPrompt):
        self.emailOut = None

        @click.command()
        def dummyCmd():
          from isitfit.utils import PromptToEmailIfNotRequested
          pte = PromptToEmailIfNotRequested()
          import tempfile
          with tempfile.NamedTemporaryFile() as fh:
            pte.last_email_cl.fn = fh.name # overwrite file to save last-used email
            # dont set to leave blank # pte.last_email_cl.set('me@example.com')
            self.emailOut = pte.prompt(emailIn)

        # https://stackoverflow.com/q/38143366/4126114
        runner = CliRunner()
        result = runner.invoke(dummyCmd, input=emailPrompt)
        return self.emailOut

    mw = MyWrap()
    actual = mw.dummyFac(None, '\n')
    assert actual is None
    actual = mw.dummyFac(None, 'n\n')
    assert actual is None
    actual = mw.dummyFac(None, 'y\nshadi@abc.com')
    assert actual == ['shadi@abc.com']
    actual = mw.dummyFac(None, 'y\nbla\nshadi@abc.com')
    assert actual == ['shadi@abc.com']





def test_word2color():
  from isitfit.utils import Word2Color
  w2c = Word2Color()
  actual = w2c.convert('whatever')
  assert actual=='green'

  actual = w2c.convert('heay')
  assert actual=='grey'

  actual = w2c.convert('another-word_under')
  assert actual=='grey'

  actual = w2c.convert('yih')
  assert actual=='cyan'


def test_pdSubsetLatest():
  import pandas as pd
  df_in = pd.DataFrame(
    {
      'a':  [1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13],
      'b':  [1, 1, 1, 2, 2, 2, 1, 1, 1, 2, 2, 2, 2],
    },
    index = [1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13]
  )
  df_expected = pd.DataFrame(
    {
      'a':  [10,11,12,13],
      'b':  [ 2, 2, 2, 2]
    },
    index = [10,11,12,13]
  )

  from isitfit.utils import pd_subset_latest
  df_actual = pd_subset_latest(df_in, 'b', 'a')

  pd.testing.assert_frame_equal(df_actual, df_expected)
  assert len(set(df_actual.b.to_list())) == 1



from isitfit.cost.ec2_analyze import Ec2Iterator



from isitfit.utils import logger




import pandas as pd
from tabulate import tabulate
import tempfile
import csv
from collections import OrderedDict

# https://pypi.org/project/termcolor/
from termcolor import colored


def df2tabulate(df):
  return tabulate(df.set_index('instance_id'), headers='keys', tablefmt='psql')


#---------------------------


def class2recommendedCore(r):
  o = { 'recommended_type': None,
        'savings': None
      }

  if r.classification_1=='Underused':
    # FIXME classification 2 will contain if it's a burstable workload or lambda-convertible
    # that would mean that the instance is downsizable twice, so maybe need to return r.type_smaller2x
    # FIXME add savings from the twice downsizing in class2recommendedType if it's a burstable workload or lambda-convertible,
    # then calculate the cost from lambda functions and add it as overhead here
    o = { 'recommended_type': r.type_smaller,
          'savings': r.cost_3m_smaller-r.cost_3m
        }

  if r.classification_1=='Idle':
    # Maybe idle servers should be recommended to "stop"
    o = { 'recommended_type': r.type_smaller,
          'savings': r.cost_3m_smaller-r.cost_3m
        }

  if r.classification_1=='Overused':
    # This is costing more
    o = {'recommended_type': r.type_larger,
         'savings': r.cost_3m_larger-r.cost_3m
        }

  return o

#---------------------------------


def ec2obj_to_name(ec2_obj):
    if ec2_obj.tags is None:
      return None

    ec2_name = [x for x in ec2_obj.tags if x['Key']=='Name']
    if len(ec2_name)==0:
      return None

    return ec2_name[0]['Value']


from isitfit.utils import taglist2str


class CalculatorOptimizeEc2:

  def __init__(self, n, thresholds = None):
    self.n = n

    if thresholds is None:
      thresholds = {
        'rightsize': {'idle': 3, 'low': 30, 'high': 70},
        'burst': {'low': 20, 'high': 80}
      }

    # iterate over all ec2 instances
    self.thresholds = thresholds
    self.ec2_classes = []

    # for csv streaming
    self.csv_fn_intermediate = None
    self.csv_fh = None
    self.csv_writer = None

  
  def __exit__(self):
    self.csv_fh.close()


  def _xxx_to_classification(self, xxx_maxmax, xxx_maxavg, xxx_avgmax):
    # check if good to convert to burstable or lambda
    # i.e. daily data shows few large spikes
    thres = self.thresholds['burst']
    if xxx_maxmax >= thres['high'] and xxx_avgmax <= thres['low'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable daily'

    # check rightsizing
    # i.e. no special spikes in daily data
    # FIXME: can check hourly data for higher precision here
    thres = self.thresholds['rightsize']
    if xxx_maxmax <= thres['idle']:
      return 'Idle', None
    elif xxx_maxmax <= thres['low']:
      return 'Underused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg >= thres['high']:
      return 'Overused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable intraday'

    return 'Normal', None


  def _ec2df_to_classification(self, ec2_df):
    # data is daily, so if less than 7 days, just return "Not enough data"
    if ec2_df.shape[0] < 7:
      return "Not enough data", "%i day(s) available. Minimum is 7 days."%ec2_df.shape[0]

    cpu_maxmax = ec2_df.cpu_used_max.max()
    cpu_maxavg = ec2_df.cpu_used_avg.max()
    cpu_avgmax = ec2_df.cpu_used_max.mean()
    cpu_c1, cpu_c2 = self._xxx_to_classification(cpu_maxmax, cpu_maxavg, cpu_avgmax)
    #print("ec2_df.{maxmax,avgmax,maxavg} = ", maxmax, avgmax, maxavg)

    if pd.isnull(ec2_df.ram_used_max).all():
      cpu_c2 = ["No ram", cpu_c2]
      cpu_c2 = [x for x in cpu_c2 if x is not None]
      cpu_c2 = ", ".join(cpu_c2)
      return cpu_c1, cpu_c2

    # continue with cpu + ram data
    ram_maxmax = ec2_df['ram_used_max'].fillna(value=0).max()
    ram_maxavg = ec2_df['ram_used_max'].fillna(value=0).mean()
    ram_avgmax = ec2_df['ram_used_avg'].fillna(value=0).max()
    ram_c1, ram_c2 = self._xxx_to_classification(ram_maxmax, ram_maxavg, ram_avgmax)

    # consolidate ram with cpu
    out_c2 = ["CPU + RAM checked",
              "CPU: %s"%cpu_c2 if cpu_c2 is not None else None,
              "RAM: %s"%ram_c2 if ram_c2 is not None else None ]
    out_c2 = ", ".join([x for x in out_c2 if x is not None])
    if cpu_c1=='Overused' or ram_c1=='Overused':
      return 'Overused', out_c2

    if cpu_c1=='Normal' or ram_c1=='Normal':
      return 'Normal', out_c2

    return 'Underused', out_c2


  def handle_pre(self, context_pre):
      # a csv file handle to which to stream results
      from isitfit.dotMan import DotMan
      self.csv_fn_intermediate = tempfile.NamedTemporaryFile(prefix='isitfit-intermediate-', suffix='.csv', delete=False, dir=DotMan().tempdir())
      import click
      click.echo(colored("Intermediate results will be streamed to %s"%self.csv_fn_intermediate.name, "cyan"))
      self.csv_fh = open(self.csv_fn_intermediate.name, 'w')
      self.csv_writer = csv.writer(self.csv_fh)

      # done
      return context_pre


  def per_ec2(self, context_ec2):
    # parse out context keys
    ec2_obj, ec2_df, mm = context_ec2['ec2_obj'], context_ec2['ec2_df'], context_ec2['mainManager']

    # filter ec2_df for the part matching the latest ec2 size only
    from isitfit.utils import pd_subset_latest
    ec2_df = ec2_df.copy()
    ec2_df = pd_subset_latest(ec2_df, 'instanceType', 'Timestamp')

    #print(ec2_obj.instance_id)
    ec2_c1, ec2_c2 = self._ec2df_to_classification(ec2_df)

    ec2_name = ec2obj_to_name(ec2_obj)

    taglist = ec2_obj.tags

    # Reported in github issue 8: NoneType object is not iterable
    # https://github.com/autofitcloud/isitfit/issues/8
    if taglist is None:
      taglist = []

    taglist = taglist2str(taglist, context_ec2['filter_tags'])

    ec2_res = OrderedDict()
    ec2_res['region'] = ec2_obj.region_name
    ec2_res['instance_id'] = ec2_obj.instance_id
    ec2_res['instance_type'] = ec2_obj.instance_type
    ec2_res['classification_1'] = ec2_c1
    ec2_res['classification_2'] = ec2_c2
    ec2_res['tags'] = taglist

    # write csv header
    if len(self.ec2_classes)==0:
      self.csv_writer.writerow(['name'] + [k for k,v in ec2_res.items() if k!='tags'])# save header

    # save intermediate result to csv file
    # Try to stick to 1 row per instance
    # Drop the tags because they're too much to include
    csv_row = [ec2_name] + [v.replace("\n", ";") for k,v in ec2_res.items() if k!='tags']
    self.csv_writer.writerow(csv_row)

    # gathering results
    self.ec2_classes.append(ec2_res)

    # check if should return early
    if self.n!=-1:
      sub_underused = [x for x in self.ec2_classes if x['classification_1']=='Underused']
      if len(sub_underused) >= self.n:
        # break early
        from isitfit.utils import IsitfitCliRunnerBreakIterator
        raise IsitfitCliRunnerBreakIterator

    # done
    return context_ec2








from isitfit.cost.base_reporter import ReporterBase

class ReporterOptimizeEc2(ReporterBase):

  def __init__(self):
    # for final csv file
    # DEPRECATED # self.csv_fn_final = None

    # members that will contain the results of the optimization
    self.df_sort = None
    self.sum_val = None


  def postprocess(self, context_all):
    # unpack
    self.analyzer = context_all['analyzer']
    self.df_cat = context_all['df_cat']

    # process
    self._after_all()
    # DEPRECATED # self._storecsv_all()

    # save to context for aggregator
    context_all['df_sort'] = self.df_sort
    context_all['sum_val'] = self.sum_val
    # DEPRECATED # context_all['csv_fn_final'] = self.csv_fn_final

    # done
    return context_all


  def _after_all(self):
    df_all = pd.DataFrame(self.analyzer.ec2_classes)

    # if no data
    if df_all.shape[0]==0:
      self.df_sort = None
      self.sum_val = None
      return

    # merge current type hourly cost
    map_cost = self.df_cat[['API Name', 'cost_hourly']]
    df_all = df_all.merge(map_cost, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge the next-smaller instance type from the catalog for instances classified as Underused
    map_smaller = self.df_cat[['API Name', 'type_smaller', 'Linux On Demand cost_smaller']].rename(columns={'Linux On Demand cost_smaller': 'cost_hourly_smaller'})
    df_all = df_all.merge(map_smaller, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge next-larger instance type
    map_larger = self.df_cat[['API Name', 'type_smaller', 'cost_hourly']].rename(columns={'type_smaller': 'API Name', 'API Name': 'type_larger', 'cost_hourly': 'cost_hourly_larger'})
    df_all = df_all.merge(map_larger, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # convert from hourly to 3-months
    for fx1, fx2 in [('cost_3m', 'cost_hourly'), ('cost_3m_smaller', 'cost_hourly_smaller'), ('cost_3m_larger', 'cost_hourly_larger')]:
      df_all[fx1] = df_all[fx2] * 24 * 30 * 3
      df_all[fx1] = df_all[fx1].fillna(value=0).astype(int)

    # imply a recommended type
    df_rec = df_all.apply(class2recommendedCore, axis=1).apply(pd.Series)
    df_all['recommended_type'], df_all['savings'] = df_rec['recommended_type'], df_rec['savings']
    df_all['savings'] = df_all.savings.fillna(value=0).astype(int)

    # keep a subset of columns
    df_all = df_all[['region', 'instance_id', 'instance_type', 'classification_1', 'classification_2', 'cost_3m', 'recommended_type', 'savings', 'tags']]

    # display
    #df_all = df_all.set_index('classification_1')
    #for v in ['Idle', 'Underused', 'Overused', 'Normal']:
    #  logger.info("\nInstance classification_1: %s"%v)
    #  if v not in df_all.index:
    #    logger.info("None")
    #  else:
    #    logger.info(df_all.loc[[v]]) # use double brackets to maintain single-row dataframes https://stackoverflow.com/a/45990057/4126114
    #
    #  logger.info("\n")

    # main results
    self.df_sort = df_all.sort_values(['savings'], ascending=True)
    self.sum_val = df_all.savings.sum()


# DEPRECATED
#  def _storecsv_all(self, *args, **kwargs):
#      if self.df_sort is None:
#        return
#
#      import tempfile
#      from isitfit.dotMan import DotMan
#      with tempfile.NamedTemporaryFile(prefix='isitfit-full-ec2-', suffix='.csv', delete=False, dir=DotMan().tempdir()) as  csv_fh_final:
#        self.csv_fn_final = csv_fh_final.name
#        logger.debug(colored("Saving final results to %s"%csv_fh_final.name, "cyan"))
#        self.df_sort.to_csv(csv_fh_final.name, index=False)
#        logger.debug(colored("Save complete", "cyan"))


# DEPRECATED
#  def display(self, context_all):
#    if self.df_sort is None:
#      logger.info(colored("No EC2 instances found", "red"))
#      return context_all
#
#    # display
#    # Edit 2019-09-25 just show the full list. Will add filtering later. This way it's less ambiguous when all instances are "Normal"
#    # self.df_sort.dropna(subset=['recommended_type'], inplace=True)
#    
#    # if no recommendations
#    if self.df_sort.shape[0]==0:
#      logger.info(colored("No optimizations from isitfit for this AWS EC2 account", "red"))
#      return context_all
#    
#    # if there are recommendations, show them
#    sum_comment = "extra cost" if self.sum_val>0 else "savings"
#    sum_color = "red" if self.sum_val>0 else "green"
#
#    import click
#    #logger.info("Optimization based on the following CPU thresholds:")
#    #logger.info(self.thresholds)
#    #logger.info("")
#    click.echo(colored("Recommended %s: %0.0f $ (over the next 3 months)"%(sum_comment, self.sum_val), sum_color))
#    click.echo("")
#
#    # display dataframe
#    from isitfit.utils import display_df
#    display_df(
#      "Recommended EC2 size changes",
#      self.df_sort,
#      self.csv_fn_final,
#      self.df_sort.shape,
#      logger
#    )
#
##    with pd.option_context("display.max_columns", 10):
##      logger.info("Details")
##      if self.df_sort.shape[0]<=10:
##        logger.info(df2tabulate(self.df_sort))
##      else:
##        logger.info(df2tabulate(self.df_sort.head(n=5)))
##        logger.info("...")
##        logger.info(df2tabulate(self.df_sort.tail(n=5)))
##        logger.info("")
##        logger.info(colored("Table originally with %i rows is truncated for top and bottom 5 only."%self.df_sort.shape[0], "cyan"))
##        logger.info(colored("Consider filtering it with --n=x for the 1st x results or --filter-tags=foo using a value from your own EC2 tags.", "cyan"))
#
#    if self.analyzer.n!=-1:
#      logger.info(colored("This table has been filtered for only the 1st %i underused results"%self.analyzer.n, "cyan"))
#
#    return context_all



def pipeline_factory(ctx, n, filter_tags):
    # moved these imports from outside the function to inside it so that `isitfit --version` wouldn't take 5 seconds due to the loading
    from isitfit.cost.mainManager import MainManager
    from isitfit.cost.cloudtrail_ec2type import CloudtrailCached

    # manager of redis-pandas caching
    from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
    cache_man = RedisPandasCacheManager()

    # 2019-12-16 deprecate direct datadog/cloudwatch listeners in favor of the automatic failover
    # from isitfit.cost.metrics_datadog import DatadogListener
    # from isitfit.cost.metrics_cloudwatch import CwEc2Listener
    from isitfit.cost.metrics_datadog import DatadogCached
    from isitfit.cost.metrics_cloudwatch import CloudwatchEc2
    from isitfit.cost.metrics_automatic import MetricsListener
    ddg = DatadogCached(cache_man)
    cloudwatchman = CloudwatchEc2(cache_man)
    metrics = MetricsListener(ddg, cloudwatchman)
    metrics.set_ndays(ctx.obj['ndays'])

    from isitfit.cost.ec2_common import Ec2TagFilter
    from isitfit.cost.catalog_ec2 import Ec2Catalog
    from isitfit.cost.ec2_common import Ec2Common

    from isitfit.tqdmman import TqdmL2Verbose
    tqdml2_obj = TqdmL2Verbose(ctx)

    ol = CalculatorOptimizeEc2(n)



    etf = Ec2TagFilter(filter_tags)


    ra = ReporterOptimizeEc2()

    mm = MainManager("EC2 cost optimize", ctx)
    mm.set_ndays(ctx.obj['ndays'])

    ec2_cat = Ec2Catalog()
    ec2_common = Ec2Common()
    ec2_it = Ec2Iterator(ctx.obj['filter_region'], tqdml2_obj)

    # boto3 cloudtrail data
    cloudtrail_manager = CloudtrailCached(mm.EndTime, cache_man, tqdml2_obj)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    inject_analyzer = lambda context_all: dict({'analyzer': ol}, **context_all)

    # utilization listeners
    mm.set_iterator(ec2_it)
    mm.add_listener('pre', cache_man.handle_pre)
    mm.add_listener('pre', cloudtrail_manager.init_data)
    mm.add_listener('pre', ol.handle_pre)
    mm.add_listener('pre', ec2_cat.handle_pre)
    mm.add_listener('ec2', etf.per_ec2)
    mm.add_listener('ec2', metrics.per_host)
    mm.add_listener('ec2', cloudtrail_manager.single)
    mm.add_listener('ec2', ec2_common._handle_ec2obj)
    mm.add_listener('ec2', ol.per_ec2)
    mm.add_listener('all', metrics.display_status)
    mm.add_listener('all', ec2_common.after_all)
    mm.add_listener('all', inject_analyzer)
    mm.add_listener('all', ra.postprocess)
    #mm.add_listener('all', ra.display)

    return mm



#import pytest
#
#
# WORK IN PROGRESS
#
#def test_issue9():
#    # prepare
#
#    # build a fake click command so that the click.prompt will be emulated
#    # https://click.palletsprojects.com/en/7.x/testing/?highlight=test#input-streams
#    import click
#    @click.command()
#    def cmd():
#      dataType = 'cost analyze (binned)'
#      import pandas as pd
#      dataVal = pd.DataFrame()
#      from isitfit.cli.core import cli_core
#      from ..emailMan import EmailMan
#      em = EmailMan(None, None, None)
#      em.send([])
#
#    # trigger
#    from click.testing import CliRunner
#    runner = CliRunner()
#    result = runner.invoke(sendemail_cmd, [], input='\n')
#    print(result.__dict__) # in case of exception, this will show details
#    assert not result.exception
#    assert True # no exception
#



from isitfit.cost.metrics_cloudwatch import CloudwatchAssistant

# Using moto yields "Error during pagination: The same next token was received twice"
# so doing the mock myself
# from moto import mock_cloudwatch

import pytest
@pytest.fixture
def mock_cloudwatch(mocker):
  mockee = 'boto3.setup_default_session'
  mocker.patch(mockee, autospec=True)

  def factory(iterator_dims, response):
    class Metric:
      def __init__(self, ndim): self.dimensions = range(ndim)
      def get_statistics(self, *args, **kwargs): return response

    class Iterator:
      def filter(self, *args, **kwargs):
        return [Metric(x) for x in iterator_dims]

    class Resource:
      metrics = Iterator()

    mockreturn = lambda *args, **kwargs: Resource()
    mockee = 'boto3.resource'
    mocker.patch(mockee, side_effect=mockreturn)

  return factory


class TestCwAssUnit:
  # @mock_cloudwatch
  def test_ok(self, mock_cloudwatch):
    import datetime as dt
    dtnow = dt.datetime.now()

    mock_cloudwatch(
      [3, 1],
      { 'Datapoints': [
          {'Timestamp': dtnow, 'SampleCount': 2, 'Maximum': 3, 'Average': 4, 'Minimum': 5}
        ]
      }
    )

    ca = CloudwatchAssistant()
    ca.set_resource("us-west-2")
    assert True # no exception

    iid = 'i-1'
    it = ca.id2iterator(iid, 'AWS/EC2', 'InstanceId')
    assert it is not None

    me = ca.iterator2metric(it, iid)
    assert me is not None

    st = ca.metric2stats(me)
    assert st is not None

    df = ca.stats2df(st, iid, dtnow, 'AWS/EC2')
    assert df is not None
    assert df.shape[0] > 0


from isitfit.cost.metrics_cloudwatch import CloudwatchEc2, CloudwatchRedshift


from isitfit.tests.cost.test_metricsDatadog import cache_man

@pytest.mark.parametrize("AdapterCls", [CloudwatchEc2, CloudwatchRedshift])
class TestCloudwatchEc2GetMetricsDerived:
  def test_yesReady_yesData(self, mocker, cache_man, AdapterCls):
    # mark as ready
    cache_man.ready = True

    # mock parent
    import pandas as pd
    mockreturn = lambda *args, **kwargs: pd.DataFrame({'a': [1,2,3]})
    mockee = 'isitfit.cost.metrics_cloudwatch.CloudwatchBase.handle_main'
    uncached_get = mocker.patch(mockee, side_effect=mockreturn)

    cwc = AdapterCls(cache_man)
    host_id = 'i-123456'

    # after first call
    actual = cwc.get_metrics_derived(None, host_id, None)
    assert actual is not None
    assert actual.shape[0]==3
    assert uncached_get.call_count == 1 # calls upstream
    assert cache_man.get.call_count == 1 # 1st check in cache
    assert cache_man.set.call_count == 1 # 1st set in cache

    # after 2nd call
    actual = cwc.get_metrics_derived(None, host_id, None)
    assert actual is not None
    assert actual.shape[0]==3
    assert uncached_get.call_count == 1 # no increment
    assert cache_man.get.call_count == 2 # 2nd check in cache
    assert cache_man.set.call_count == 1 # no increment



"""
General tests for click usage. Check each test's docstrings for more details
"""

import click

class TestClickOptionMultipleWithDefault:
  """
  General tests for click to check how it deals with defaults and multiple=True (for share-email option)
  The goal is to figure out when to *not* prompt the user for an email even if it is not passed
  In other words, finding a signal about the difference between --share-email empty because not passed or --share-email empty because intended

  Conclusion: click doesn't offer such a signal without adding a separate option --skip-share-email
  """
  def test_multipleWithDefault(self):
    # do not set default, do not pass on CLI
    @click.command()
    @click.option('--share-email', multiple=True, help='Share result to email address')
    def doit_defaultNo(share_email):
      assert type(share_email) is tuple
      assert len(share_email)==0

    from click.testing import CliRunner
    runner = CliRunner()

    result = runner.invoke(doit_defaultNo, [], input='\n')
    assert not result.exception


    # set default to None, do not pass on CLI
    @click.command()
    @click.option('--share-email', multiple=True, help='Share result to email address', default=None)
    def doit_defaultYes(share_email):
      assert share_email is not None # despite the default
      assert type(share_email) is tuple
      assert len(share_email)==0

    from click.testing import CliRunner
    runner = CliRunner()

    result = runner.invoke(doit_defaultYes, [], input='\n')
    assert not result.exception


    # set default to None, pass some input on CLI
    e1, e2 = 'e1@foo.com', 'e2@bla.com'

    # set default to None, pass arguments on CLI
    @click.command()
    @click.option('--share-email', multiple=True, help='Share result to email address', default=None)
    def doit_defaultYes(share_email):
      assert type(share_email) is tuple
      assert len(share_email)==2
      assert share_email == (e1, e2)

    from click.testing import CliRunner
    runner = CliRunner()

    result = runner.invoke(doit_defaultYes, [ f'--share-email={e1}', f'--share-email={e2}' ], input='\n')
    assert not result.exception


    # do not set default, do not pass on CLI, add option to skip
    @click.command()
    @click.option('--share-email', multiple=True, help='Share result to email address')
    @click.option('--skip-prompt-email', is_flag=True, help='Skip the prompt to share result to email address')
    def doit_defaultNo(share_email, skip_prompt_email):
      assert type(share_email) is tuple
      assert len(share_email)==0
      assert skip_prompt_email

    from click.testing import CliRunner
    runner = CliRunner()

    result = runner.invoke(doit_defaultNo, ['--skip-prompt-email'], input='\n')
    assert not result.exception


